/*
 * Barriers.java
 *
 * Created on 29. srpen 2003, 11:34
 */
import java.lang.IllegalArgumentException;
import java.util.*;
import java.awt.*;
/** Class representing barriers in a grid.
 * @author Pavel
 */
public class Barriers {
    
    private int width;
    private int height;
    private int positionX;
    private int positionY;
    private int numberOfBarriers;
    private boolean[][] barriersGrid;
    private Color barriersColor = Color.black;
    
    private Random random;
    
    /** Creates a new instance of Barriers
     * @param width Width of grid where to create barriers. (Number of cells.)
     * @param height Height of grid where to create barriers. (Number of cells.)
     * @param numberOfBarriers Number of barriers to create. Must be 0 <= numberOfBarriers < width * height *
     * 10
     * @param positionX X-coordinate of upper-left corner of a grid where barriers are displayed (in
     * pixels).
     * @param positionY Y-coordinate of upper-left corner of a grid where barriers are displayed (in
     * pixels).
     * @param headPositionX X-coordinate of a point in grid in whose nearness are no barriers created. In
     * the concrete it's 3 cells around that point. Usually this is a position of
     * caterpillar's head.
     * @param headPositionY Y-coordinate of a point in grid in whose nearness are no barriers created. In
     * the concrete it's 3 cells around that point. Usually this is a position of
     * caterpillar's head.
     * @throws IllegalArgumentException If too big number of barriers is passed, <CODE>IllegalArgumentException</CODE>
     * is thrown. Too big means greather then (width * height / 10).
     */
    public Barriers(int width, int height, int numberOfBarriers, int positionX, int positionY,
                int headPositionX, int headPositionY) throws IllegalArgumentException, RuntimeException
    {
        if (numberOfBarriers > (width * height / 10))
            throw new IllegalArgumentException("numberOfBarriers to big");
        if (numberOfBarriers < 0)
            throw new IllegalArgumentException("numberOfBarriers is less than 0");
        if (width < 0)
            throw new IllegalArgumentException("width is less than 0");
        if (height < 0)
            throw new IllegalArgumentException("height is less than 0");
        
        this.width = width;
        this.height = height;
        this.numberOfBarriers = numberOfBarriers;
        this.positionX = positionX;
        this.positionY = positionY;
        random = new Random();
        
        barriersGrid = new boolean[width][height];
        clearBarriersGrid();
        setBarriersGrid(headPositionX, headPositionY);
    }
    
    private void clearBarriersGrid()
    {
        int i, j;
        for (i = 0; i < width; i++)
            for (j = 0; j < height; j++)
                barriersGrid[i][j] = false;
    }
    
    private void setBarriersGrid(int headPositionX, int headPositionY) throws RuntimeException
    {
        int barriersSet = 0;
        int i, j;
        int xHeadDistance, yHeadDistance;
        int counter = 0; // to stop perpetual cycle
        int stop = numberOfBarriers * 10000;
        
        while (barriersSet < numberOfBarriers)
        {
            if ((counter++ >= Integer.MAX_VALUE) || (counter > stop))
                throw new java.lang.RuntimeException("Probably barriers could not be set.");
            i = random.nextInt(width);
            j = random.nextInt(height);
            
            xHeadDistance = i - headPositionX;
            yHeadDistance = j - headPositionY;
            if (xHeadDistance < 0)
                xHeadDistance = -1 * xHeadDistance;
            if (yHeadDistance < 0)
                yHeadDistance = -1 * yHeadDistance;
            
            if ((barriersGrid[i][j] == false) && ((xHeadDistance > 3) || (yHeadDistance > 3)))
            {
                barriersGrid[i][j] = true;
                barriersSet++;
            }
        }
    }
    
    /** Tests if there is a barrier at the specific point.
     * @param x X-coordinate (in cells)
     * @param y Y-coordinate (in cells)
     * @return True if there is a barrier, false otherwise.
     */    
    public boolean IsBarrier(int x, int y)
    {
        return barriersGrid[x][y];
    }
    
    private Point countCellPosition(int cellX, int cellY)
    {
        return Grid.countCellPosition(cellX, cellY, positionX, positionY);
    }
    
    /** Paints barriers in a form.
     * @param g Graphics object used to painting.
     */    
    public void paint(Graphics g)
    {
        g.setColor(barriersColor);
        
        int i, j;
        Point p;
        for (i = 0; i < width; i++)
            for (j = 0; j < height; j++)
            {
                if (barriersGrid[i][j] == true)
                {
                    p = countCellPosition(i, j);
                    g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
                }
            }
    }
    
    /** Clears barriers from a grid on a form. (Repaints the cells by the background
     * color.)
     * @param g Graphics object used to painting.
     * @param backgroundColor Background color.
     */    
    public void Delete(Graphics g, Color backgroundColor)
    {
        g.setColor(backgroundColor);
        
        Point p;
        int i, j;
        for (i = 0; i < width; i++)
            for (j = 0; j < height; j++)
                if (barriersGrid[i][j] == true)
                {
                    p = countCellPosition(i, j);
                    g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
                }
    }
    
}
