/*
 * Caterpillar.java
 *
 * Created on 27. srpen 2003, 13:48
 */
import java.util.*;
import java.awt.*;
/** Caterpillar class. Represents the caterpillar and barriers in the grid where the
 * caterpillar is moving.
 * @author Pavel
 */
public class Caterpillar {

    private class CellPoint{
        public int x;
        public int y;
        public CellPoint next;
        public CellPoint previous;
    }
    
    private class MovingRequest
    {
        public int whereToMove;
        public MovingRequest next;
    }
    
    private MovingRequest firstMovingRequest;
    private MovingRequest lastMovingRequest;
    
    /** Width - number of cells.
     */
    private int width;
    /** Height - number of cells.
     */
    private int height;
    /** X coordinate of grid where caterpillar moves
     */
    private int positionX;
    /** Y coordinate of grid where caterpillar moves
     */
    private int positionY;
    
    /** true where is body of caterpillar except head
     */
    private boolean[][] caterpillarCells;
    private int caterpillarLength = 3;
    
    /** Caterpillar can be shorter then <code>caterpillarLength</code> when moving
     * from the border of grid when starting game.
     */
    private int caterpillarCurrentLength;
    
    private static Color caterpillarBodyColor = Color.red;
    private static Color caterpillarHeadColor = Color.blue;
    private static Color finalPointColor = Color.green;
    private Color backgroundColor;
    
    /** Grid where is this caterpillar moving
     */
    private Grid grid;
    
    private Barriers barriers;
    
    /** Constant representing caterpillar moving direction from left to right. */    
    public static final int TO_RIGHT = 1;
    /** Constant representing caterpillar moving direction from right to left. */    
    public static final int TO_LEFT = 2;
    /** Constant representing caterpillar moving direction downwards. */    
    public static final int TO_DOWN = 3;
    /** Constant representing caterpillar moving direction upwards. */    
    public static final int TO_UP = 4;
    
    /** Direction to where is caterpillar currently moving.
     */
    private int currentlyMoving;
    
    private CellPoint head;
    private CellPoint tail;
    
    private Random random;
    
    private int finalPointX = -1;
    private int finalPointY = -1;
    
    /** Creates a new instance of Caterpillar
     * @param width Width of grid where caterpillar can move. (Number of cells.)
     * @param height Height of grid where caterpillar can move. (Number of cells.)
     * @param positionX x-coordinate in pixels of upper-left corner of a grid.
     * @param positionY y-coordinate in pixels of upper-left corner of a grid.
     * @param backgroundColor Background color of a form.
     * @param grid Refererence to Grid object instance where caterpillar is moving.
     * @param numberOfBarriers Number of barriers that will be created in a grid.
     * @param g Graphics object that is used to paint caterpillar and barriers.
     */
    public Caterpillar(int width, int height, int positionX, int positionY,
            Color backgroundColor, Grid grid, int numberOfBarriers, Graphics g)
    {
        this.width = width;
        this.height = height;
        this.positionX = positionX;
        this.positionY = positionY;
        this.backgroundColor = backgroundColor;
        this.grid = grid;
        caterpillarCells = new boolean[width][height];
        for (int i = 0; i < width; i++)
            for (int j = 0; j < height; j++)
                caterpillarCells[i][j] = false;
        
        random = new Random();
        currentlyMoving = random.nextInt(4) + 1;
        
        head = new CellPoint();
        switch (currentlyMoving)
        {
            case TO_RIGHT: // right
                head.x = 0;
                head.y = random.nextInt(height);
                break;
            case TO_LEFT:
                head.x = width - 1;
                head.y = random.nextInt(height);
                break;
            case TO_DOWN:
                head.x = random.nextInt(width);
                head.y = 0;
                break;
            case TO_UP:
                head.x = random.nextInt(width);
                head.y = height - 1;
                break;
        }
        head.next = null;
        head.previous = null;
        tail = head;
        caterpillarCurrentLength = 1;
        
        barriers = new Barriers(grid.GetWidth(), grid.GetHeight(), numberOfBarriers, positionX, positionY, head.x, head.y);
        createFinalPoint(g);
    }
    
    /** Repaints whole caterpillar and all barriers.
     * @param g Graphics object that is used to painting.
     */    
    public void paint(Graphics g)
    {
        Point p;
        g.setColor(finalPointColor);
        p = countCellPosition(finalPointX, finalPointY);
        g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
        
        if (head != null)
        {
            g.setColor(caterpillarHeadColor);
            if (IsCorrectHeadPosition()) // null means correct
            {
                p = countCellPosition(head.x, head.y);
                g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
            }

            g.setColor(caterpillarBodyColor);
            CellPoint cp = head.next;
            while ((cp != tail) && (cp != null))
            {
                p = countCellPosition(cp.x, cp.y);
                g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
                cp = cp.next;
            }
        
            if (cp == tail)
            {
                p = countCellPosition(cp.x, cp.y);
                g.fillRect(p.x + 3, p.y + 3, Grid.cellSize - 6, Grid.cellSize - 6);
            }
        }
        barriers.paint(g);
    }
    
    private void paintHead(Graphics g)
    {
        g.setColor(caterpillarHeadColor);

        if (head == null)
            return;
        
        Point p;
        if (IsCorrectHeadPosition()) // null means correct
        {
            p = countCellPosition(head.x, head.y);
            g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
        }
        if (head.next != null)
        {
            g.setColor(caterpillarBodyColor);
            p = countCellPosition(head.next.x, head.next.y);
            g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
        }
    }

    private void clearTail(Graphics g)
    {
        g.setColor(backgroundColor);
        if (tail == null)
            return;
        Point p = countCellPosition(tail.x, tail.y);
        g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
        
        if ((tail.previous != null) && (tail.previous != head))
        {
            p = countCellPosition(tail.previous.x, tail.previous.y);
            g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
            g.setColor(caterpillarBodyColor);
            g.fillRect(p.x + 3, p.y + 3, Grid.cellSize - 6, Grid.cellSize - 6);
        }
    }

    private Point countCellPosition(int cellX, int cellY)
    {
        return Grid.countCellPosition(cellX, cellY, positionX, positionY);
    }
    
    private void ProcessMovingRequest()
    {
        if (firstMovingRequest == null)
            return;
        currentlyMoving = firstMovingRequest.whereToMove;
        
        if (firstMovingRequest == lastMovingRequest)
            lastMovingRequest = null;
        
        firstMovingRequest = firstMovingRequest.next;
    }
    
    /** Moves caterpillar's body one cell in direction that has been set by force of
     * <CODE>SetMoveDirection(int)</CODE>. If this movement causes that head is on a
     * barrier or not in a grid, head isn't painted.
     * If is not correct head position, <CODE>Move(Graphics)</CODE> does nothing.
     * @param g Graphics object that is used to repaint caterpillar's position
     */    
    public void Move(Graphics g)
    {
        if (IsCorrectHeadPosition())
        {
            ProcessMovingRequest();
            MoveHead(g);
            TestFinalPoint(g);
            MoveTail(g);
        }
    }
    
    private void MoveHead(Graphics g)
    {
        caterpillarCells[head.x][head.y] = true;
        CellPoint newHead = new CellPoint();
        newHead.next = head;
        head.previous = newHead;
        newHead.previous = null;
        switch (currentlyMoving)
        {
            case TO_RIGHT: // right
                newHead.x = head.x + 1;
                newHead.y = head.y;
                break;
            case TO_LEFT:
                newHead.x = head.x - 1;
                newHead.y = head.y;
                break;
            case TO_DOWN:
                newHead.x = head.x;
                newHead.y = head.y + 1;
                break;
            case TO_UP:
                newHead.x = head.x;
                newHead.y = head.y - 1;
                break;
        }
        head = newHead;
        
        caterpillarCurrentLength++;
        
        paintHead(g);
    }
    
    private void MoveTail(Graphics g)
    {
        while (caterpillarCurrentLength > caterpillarLength)
        {
            clearTail(g);
            caterpillarCells[tail.x][tail.y] = false;
            tail = tail.previous;
            tail.next = null;
            
            caterpillarCurrentLength--;
        }
    }
    
    private void TestFinalPoint(Graphics g)
    {
        if (IsAtFinalPoint())
        {
            caterpillarLength++;
            createFinalPoint(g);
        }
    }
    
    /** Tests if caterpillar's head is on a barrier or not in a grid (incorrect
     * position) or is somewhere else in a grid (correct position).
     * @return True if head has correct position, false otherwise.
     */    
    public boolean IsCorrectHeadPosition()
    {
        if (head != null)
        {
            if (!IsCorrectPosition(head.x, head.y))
                return false;

            if (caterpillarCells[head.x][head.y] == true)
                // crashed into herself?
                return false;
        }
      
        return true;
    }
    
    private boolean IsCorrectPosition(int x, int y)
    {
        if ((x < 0) || (x >= width) || (y < 0) || (y >= height))
            // is out of grid?
            return false;
        if (barriers.IsBarrier(x, y))
            return false;
        
        return true;
    }
    
    private boolean IsAtFinalPoint()
    {
        if ((head.x == finalPointX) && (head.y == finalPointY))
            return true;
        
        return false;
    }
    
    /** Call this method to remove caterpillar from grid correctly.
     * Deletes caterpillar from a form. Background color passed to constructor is
     * used.
     * @param g Graphics object used to painting.
     */
    public void Delete(Graphics g)
    {
        g.setColor(backgroundColor);
        
        Point p;
        int i, j;
        for (i = 0; i < width; i++)
            for (j = 0; j < height; j++)
                if (caterpillarCells[i][j] == true)
                {
                    p = countCellPosition(i, j);
                    g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
                }
        
        if (IsCorrectHeadPosition())
        {
            p = countCellPosition(head.x, head.y);
            g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
        }          
        
        barriers.Delete(g, backgroundColor);
        
        p = countCellPosition(finalPointX, finalPointY);
        g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
    }
    
    /** Sets future moving direction of caterpillar.
     * @param direction Direction. Use constants provided by this class (TO_RIGHT etc.)
     */    
    public void SetMoveDirection(int direction)
    {
        int lastMoving;
        int newMoving = 0; //invalid value
        
        if (lastMovingRequest != null)
            lastMoving = lastMovingRequest.whereToMove;
        else
            lastMoving = currentlyMoving;
        
        switch (lastMoving)
        {
            case TO_RIGHT:
            case TO_LEFT:
                if ((direction == TO_UP) || (direction == TO_DOWN))
                {
                    newMoving = direction;
                }
                break;
            case TO_DOWN:
            case TO_UP:
                if ((direction == TO_RIGHT) || (direction == TO_LEFT))
                {
                    newMoving = direction;
                }
                break;
        }
        
        if (newMoving == 0)
            return;
        
        if (lastMovingRequest == null)
        {
            firstMovingRequest = new MovingRequest();
            firstMovingRequest.whereToMove = newMoving;
            firstMovingRequest.next = null;
            lastMovingRequest = firstMovingRequest;
        }
        else
        {
            lastMovingRequest.next = new MovingRequest();
            lastMovingRequest = lastMovingRequest.next;
            lastMovingRequest.next = null;
            lastMovingRequest.whereToMove = newMoving;
        }
    }

    private void createFinalPoint(Graphics g)
    {
        int xHeadDistance, yHeadDistance;
        int i, j;
        
        do
        {
            finalPointX = random.nextInt(width);
            finalPointY = random.nextInt(height);
            
            xHeadDistance = finalPointX - head.x;
            yHeadDistance = finalPointY - head.y;
            if (xHeadDistance < 0)
                xHeadDistance = -1 * xHeadDistance;
            if (yHeadDistance < 0)
                yHeadDistance = -1 * yHeadDistance;
        }
        while(((xHeadDistance <= 3) && (yHeadDistance <= 3))
            || barriers.IsBarrier(finalPointX, finalPointY)
            || caterpillarCells[finalPointX][finalPointY]
            || !pointExistPath(finalPointX, finalPointY)
            || !pointExistTwoPaths(finalPointX, finalPointY));
        
        g.setColor(finalPointColor);
        Point p = countCellPosition(finalPointX, finalPointY);
        g.fillRect(p.x + 1, p.y + 1, Grid.cellSize - 2, Grid.cellSize - 2);
    }
    
    /** Return true if there is path from head to [x,y]
     */
    private boolean pointExistPath(int x, int y)
    {
        boolean[][] reachablePoints = new boolean[width][height];
        
        int i,j;
        for (i = 0; i < width; i++)
            for (j = 0; j < height; j++)
                reachablePoints[i][j] = false;
        
        Stack toProcess = new Stack();
        reachablePoints[head.x][head.y] = true;
        toProcess.push(new Point(head.x, head.y));
        
        Point p;
        int situation;
        int newX = 0, newY = 0; // initialization for compiler satisfaction
        while (!toProcess.isEmpty())
        {
            p = (Point) toProcess.pop();
            
            for (situation = 0; situation < 4; situation++)
            {
                switch (situation)
                {
                    case 0: newX = p.x - 1; newY = p.y; break;
                    case 1: newX = p.x + 1; newY = p.y; break;
                    case 2: newX = p.x; newY = p.y - 1; break;
                    case 3: newX = p.x; newY = p.y + 1; break;
                }
                
                if (IsCorrectPosition(newX, newY) && !reachablePoints[newX][newY])
                {
                    reachablePoints[newX][newY] = true;
                    toProcess.push(new Point(newX, newY));
                }
            }
            
            if (reachablePoints[x][y])
                return true;
        }
        
        return false;
    }
    
    private boolean pointExistTwoPaths(int x, int y)
    {
        int numOfPaths = 0;
        if (IsCorrectPosition(x - 1, y)) numOfPaths++;
        if (IsCorrectPosition(x + 1, y)) numOfPaths++;
        if (IsCorrectPosition(x, y - 1)) numOfPaths++;
        if (IsCorrectPosition(x, y + 1)) numOfPaths++;
        
        if (numOfPaths >= 2)
            return true;
        
        return false;
    }
}
